(* Main *)
(* $Id: main.ml,v 1.3 2002/09/02 18:04:11 durak Exp $ *)
(* By Berke Durak *)
(* Thanks to Hugo Gimbert *)

open Unix
open Bigarray
open Debug

module IM = Map.Make(struct type t = int let compare = compare end)
module IS = Set.Make(struct type t = int let compare = compare end)
module IIS = Set.Make(struct type t = int * int let compare = compare end)
module IIM = Map.Make(struct type t = int * int let compare = compare end)

module Sliding :
	sig
	  type 'a t
	  val create : int -> 'a -> 'a t
	  val clear : 'a t -> unit
	  val push : 'a t -> 'a -> unit
	  val length : 'a t -> int
	  val get : 'a t -> int -> 'a
	  val fold_some : ('b -> 'a -> 'b) -> 'b -> 'a t -> int -> 'b
	  val fold : ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b
	end
	=
  struct
	type 'a t = {
		m : int; (* length of array *)
		mutable n : int; (* number of elements *)
		mutable i : int; (* index of _last_ element *)
		a : 'a array;
		x : 'a (* initial element *)
	  }
	let clear t =
	  t.n <- 0;
	  t.i <- -1;
	  for i = 0 to t.m - 1 do
		t.a.(i) <- t.x
	  done
	let fold_some f q0 t n =
	  if n < 0 then
		invalid_arg "Sliding.fold"
	  else
		let n = min n t.n in
		let rec loop i j q =
		  if i = n then
			q
		  else
			loop
			  (i + 1)
			  (if j = 0 then t.m - 1 else j - 1)
			  (f q (t.a.(j)))
		in
		loop 0 t.i q0
	let fold f q0 t = fold_some f q0 t t.n
	let create m x =
	  { m = m;
		n = 0;
		i = -1;
		a = Array.create m x;
		x = x }
	let push t x =
	  if t.n < t.m then
		begin
		  t.i <- 1 + t.i;
		  t.a.(t.i) <- x;
		  t.n <- 1 + t.n
		end
	  else
		begin
		  t.i <- if t.i + 1 < t.m then t.i + 1 else 0;
		  t.a.(t.i) <- x
		end
	let length t =  t.n
	(* 0 = last element *)
	(* 1 = previous element ... *)
	let get t i =
	  if i < 0 or i >= t.n then
		invalid_arg "Sliding.get"
	  else
		let j = t.i - i in
		if j < 0 then
		  t.a.(j + t.m)
		else
		  t.a.(j)
  end

type 'a knowledge = Unknown | Known of 'a
type tile = Plain | Water | Wall | Base
type base = {
	bi : int;
	bj : int;
	mutable bpackages : int knowledge; (* estimated number of packages *)
  }
type package = {
	pid : int;
	mutable ppickup : int;
	mutable powner : int option;
	mutable pi : int;
	mutable pj : int;
	mutable pdi : int;
	mutable pdj : int;
	mutable pw : int (* knowledge *)
  }
type robot = {
	id : int;
	mutable i : int;
	mutable j : int;
	mutable alive : bool;
	mutable carrying : IS.t;
	history : (int * int) Sliding.t;
	mutable avoid : int option;
  }

type predictor = {
	mutable score : float; (* the score of this prediction function *)
	predictor : (robot -> (int * int) list); (* the prediction function *)
  }	
type robot_tracking = {
	robot : robot;
	predictors : predictor list;
	mutable last_track : int
  }

type direction = East | West | North | South
type event =
	Moved of direction |
	Picked of int | Dropped of int | Appeared of int * int
type command = Move of direction | Pick of int list | Drop of int list

module Opt =
  struct
	let crt = ref false
	let crt_tile_width = ref 16
	let crt_tile_height = ref 16
	let crt_max_width = ref 350
	let crt_max_height = ref 500
	let crt_font = ref "-*-clean-medium-r-normal--6-*-*-*-c-40-*-*"
	let max_package_tries = ref 100
	let randomization = ref 0.1
	let robot_history_length = ref 16
	let robot_radius = ref 10
	let robot_tracking_persistence = ref 10
	(* let occupation_update_rate = ref 0.99 *)
	let occupation_threshold = ref 0.5
	let angriness_avoidance_threshold = ref 5
	let specs = [
	  "-crt",Arg.Set(crt),"display status in graphical window";
	  "-crt-font",Arg.String((:=)crt_font),"font to use to draw robot id's";
	  "-crt-tile-width",Arg.Int((:=) crt_tile_width),"tile width for displaying";
	  "-crt-tile-height",Arg.Int((:=) crt_tile_height),"tile height for displaying";
	  "-crt-max-width",Arg.Int((:=) crt_max_width),"maximum window width before reducing size of tiles";
	  "-crt-max-height",Arg.Int((:=) crt_max_height),"maximum window height before reducing size of tiles";
	  "-randomization",Arg.Float((:=) randomization),"randomization value (0.1)";
	  "-max-package-tries",Arg.Int((:=) max_package_tries),"maximum number of tries for bin packing";
	  "-robot-history-length",Arg.Int((:=) robot_history_length),"robot history length (10)";
	  "-robot-radius",Arg.Int((:=) robot_radius),"robot radius (10)";
	  "-robot-tracking-persistence",Arg.Int((:=) robot_radius),"robot tracking persistance in rounds (10)";
	  (* "-occupation-update-rate",Arg.Float((:=) occupation_update_rate),"robot occupation update rate (0.99)"; *)
	  "-occupation-threshold",Arg.Float((:=) occupation_threshold),"robot occupation threshold (0.5)";
	  "-debug",Arg.Int(fun x -> debug_enabled := (Only[Debug x],All)::!debug_enabled),"enable debugging";
	  "-trace",Arg.Int(fun x -> debug_enabled := (Only[Trace x],All)::!debug_enabled),"enable tracing level";
	  "-angriness-avoidance-threshold",Arg.Int((:=) angriness_avoidance_threshold),"angriness avoidance threshold";
	]
  end

module Crt =
  struct
	type displayon =
		Package of int
	  | Robot of int
	  | Dead_robot of int
	  | Target of int
	  | Delivery
	  | Myself
	  | Trail
	  |	Occupied
	  |	Track
	  |	Prediction
	  |	Pseudo_wall
	type element = {
		death : float; (* time of death *)
		what : displayon
	  }
	module IIIM = Map.Make(struct type t = int * int * int let compare = compare end)
	open Graphics

	type ops = {
		create : (int -> int -> displayon -> float -> int * int * int);
		move : (int * int * int -> int -> int -> unit);
		remove : (int * int * int -> unit);
		redisplay : (tile array array -> (int * int * displayon) list -> unit);
	  }

	let dummy_ops = {
	  create = (fun _ _ _ _ -> (0,0,0));
	  move = (fun _ _ _ -> ());
	  remove = (fun _ -> ());
	  redisplay = (fun _ _ -> ())
	}

	let init m n =
	  let serial = ref 0 in
	  let things = ref IIIM.empty in
	  let p,q = !Opt.crt_tile_height,!Opt.crt_tile_width in
	  if m > !Opt.crt_max_height or n > !Opt.crt_max_width then
		failwith "map too big"
	  else
		if m = 0 or n = 0 then
		  failwith "null-sized map"
		else
		  let p,q =
			if m * p <= !Opt.crt_max_height && n * q <= !Opt.crt_max_width then
			  p,q
			else
			  let pq = min (!Opt.crt_max_height/m) (!Opt.crt_max_width/n) in
			  pq,pq
		  in
		  open_graph (Printf.sprintf " %dx%d+0+0" (n * q) (m * p));
		  let plain_color = rgb 0x0 0xff 0x00
		  and water_color = rgb 0x00 0x00 0xff
		  and wall_color = rgb 0x80 0x80 0x80
		  and base_color = rgb 0xff 0xff 0x00
		  and robot_color = rgb 0x00 0x00 0x00
		  and myself_color = rgb 0x00 0x00 0x00
		  and dead_robot_color = rgb 0x00 0x00 0xff
		  and package_color = rgb 0xff 0x00 0x00
		  and occupied_color = rgb 0xff 0x00 0x00
		  and target_color = rgb 0xff 0x80 0x00
		  and delivery_color = rgb 0xff 0x80 0x80
		  and trail_color = rgb 0xff 0xff 0xff
		  and track_color = rgb 0xff 0x00 0x00
		  and prediction_color = rgb 0xff 0xff 0x00
		  and pseudo_wall_color = rgb 0x00 0x00 0x00
		  in
		  auto_synchronize false;
		  set_font !Opt.crt_font;
		  let last_map = Array.make_matrix m n None in
		  (* create *)
		  {
		   create = (fun i j what lifetime ->
			 incr serial;
			 things :=
			   IIIM.add (!serial,i,j)
				 { death =
				   (if lifetime > 0.0 then
					 Unix.gettimeofday () +. lifetime
				   else
					 0.0);
				   what = what }
				 !things;
			 (!serial,i,j));
		   (* move *)
		   move = (fun (serial,i,j) i' j' ->
			 try
			   let e = IIIM.find (serial,i,j) !things in
			   things :=
				 IIIM.add (serial,i',j') e
				   (IIIM.remove (serial,i,j) !things)
			 with
			   Not_found -> debug (Warning,System) (sf "graphic object (%d,%d,%d) not found" serial i j));
		   (* remove *)
		   remove = (fun (serial,i,j) -> things := IIIM.remove (serial,i,j) !things);
		   (* redisplay *)
		   redisplay = (fun a temporary ->
			 let t = Unix.gettimeofday () in
			 IIIM.iter
			   (fun (id,i,j) e ->
				 if e.death != 0.0 && e.death < t then
				   begin
					 last_map.(i).(j) <- None; (* must update this *)
					 things := IIIM.remove (id,i,j) !things
				   end) !things;
			 for i = 0 to m - 1 do
			   for j = 0 to n - 1 do
				 if last_map.(i).(j) <> Some a.(i).(j) then
				   begin
					 set_color
					   (match a.(i).(j) with
						 Plain -> plain_color
					   | Water -> water_color
					   | Wall -> wall_color
					   | Base -> base_color);
					 fill_rect (j * q) ((m - 1 - i) * p) q p;
					 last_map.(i).(j) <- Some a.(i).(j)
				   end
			   done
			 done;
			 let draw i j what =
			   if i >= 0 && i < m && j >= 0 && j < n then
				 begin
				   last_map.(i).(j) <- None;
				   moveto (j * q) ((m - 1 - i) * p);
				   match what with
					 Package x ->
					   set_color package_color;
					   draw_string (sf "%x" x)
				   | Robot x ->
					   set_color robot_color;
					   draw_string (sf "%x" x)
				   | Dead_robot x ->
					   set_color dead_robot_color;
					   draw_string (sf "%x" x)
				   | Target x ->
					   set_color target_color;
					   draw_string (sf "%x" x)
				   | Delivery ->
					   set_color myself_color;
					   draw_string "##"
				   | Myself ->
					   set_color myself_color;
					   draw_string "O"
				   | Trail ->
					   set_color trail_color;
					   draw_string "#"
				   | Occupied ->
					   set_color occupied_color;
					   draw_string "!!!"
				   | Track ->
					   set_color track_color;
					   draw_rect (j * q) ((m - 1 - i) * p + 1) (p - 1) (q - 1)
				   | Prediction ->
					   set_color prediction_color;
					   draw_rect (j * q + 1) ((m - 1 - i) * p + 2) (p - 3) (q - 3)
				   | Pseudo_wall ->
					   set_color pseudo_wall_color;
					   draw_rect (j * q + 2) ((m - 1 - i) * p + 3) (p - 4) (q - 4)

				 end
			 in
			 IIIM.iter (fun (id,i,j) e -> draw i j e.what) !things;
			 List.iter (fun (i,j,what) -> draw i j what) temporary;
			 synchronize ()) }
  end

let split_at c s =
  let m = String.length s in
  let i = String.index s c in
  (String.sub s 0 i,
   String.sub s (i + 1) (m - i - 1))
	
let split_multiple_at_char c b u =
  let m = String.length u in
  let rec loop0 r i =
    if i >= m then
      List.rev r
    else
      if u.[i] = c then
        loop0 r (i + 1)
      else
        loop1 r i
  and loop1 r i =
    if i = m or u.[i] = c then
      begin
        let x = Buffer.contents b in
        Buffer.clear b;
        loop0 (x::r) (i + 1)
      end
    else
      begin
        Buffer.add_char b u.[i];
        loop1 r (i + 1)
      end
  in
  Buffer.clear b;
  loop0 [] 0

let int_of_string x =
  try
	int_of_string x
  with
	_ -> failwith (sf "int_of_string failed on %S" x)

(* y : 1...m *)
(* i : 0 ... m - 1 *)
(* i = m - y *)
(* j = x - 1 *)

module Bitmap :
	sig
	  type t
	  val make : int -> bool -> t
	  val clear : t -> unit
	  val set : t -> int -> bool -> unit
	  val get : t -> int -> bool
	end
	=
  struct
	type t = int * string
	let make m x = (m,String.make ((m + 7) lsr 3) (if x then '\255' else '\000'))
	let clear (m,s) = String.fill s 0 (String.length s) '\000'
	let set (m,s) i x =
	  assert (0 <= i && i < m);
	  if x then
		s.[i lsr 3] <- Char.chr ((1 lsl (i land 7)) lor Char.code s.[i lsr 3])
	  else
		s.[i lsr 3] <- Char.chr ((lnot (1 lsl (i land 7))) land Char.code s.[i lsr 3])
	let get (m,s) i =
	  assert (0 <= i && i < m);
	  0 <> (1 lsl (i land 7)) land Char.code s.[i lsr 3]
  end

let two_pi = 6.283185307179586232

let pack_position (i,j) =
  assert (0 <= i && i < 16384 && 0 <= j && j < 16384);
  (i lsl 14) lor j

let unpack_position x = ((x lsr 14),(x land 16383))

let swap a i j =
  let x = a.(i) in
  a.(i) <- a.(j);
  a.(j) <- x

let scramble a =
  let m = Array.length a in
  for i = 0 to m - 1 do
    swap a i (i + (Random.int (m - i)))
  done

module Breadth_first =
  struct
	(* a : map, m x n tile array array *)
	(* s : source int * int *)
	(* t : target (int * int) list *)
	let directions =
	  let a = [|1,0;0,1;-1,0;0,-1|] in
	  Array.init 16 (fun _ ->
		scramble a;
		Array.to_list a)

	let distances m n =
	  let b = Bitmap.make (m * n) false in
	  fun walkable (si,sj) t ->
		(* debug (Trace 5,Game) (sf "breadth first search from (%d,%d) to [%s]" si sj
		   (String.concat ";" (List.map (fun (i,j) -> sf "(%d,%d)" i j) t))); *)
		Bitmap.clear b;
		let is_marked i j = Bitmap.get b (n * i + j)
		and mark i j = Bitmap.set b (n * i + j) true
		in
		(* d : current distance *)
		(* q : contents of current queue ; invariant : distance from s to all elemnts of q is d *)
		(* r : results *)
		let rec loop d q res =
		  (* process ring d ; invariant : all elements of q are marked and have distance d *)
		  (* debug (Trace 10,Game) (sf "ring %d" d); *)
		  match q with
			[] -> res
		  | _ ->
			  (* process ring, collecting new neighbours in q' *)
			  (* and updating results in res *)
			  let rec loop2 q' res lst =
				match lst with
				  (i,j)::r ->
					(* debug (Trace 10,Game) (sf "considering neighbours of (%d,%d)" i j); *)
					loop2
					  (List.fold_left
						 (fun l (di,dj) ->
						   let i' = i + di
						   and j' = j + dj
						   in
						   if i' >= 0 && i' < m && j' >= 0 &&  j' < n &&
							 not (is_marked i' j') &&
							 walkable i' j' then
							 begin
							   (* debug (Trace 10,Game) (sf "neighbour (%d,%d) is appropriate" i' j'); *)
							   mark i' j';
							   (i',j')::l
							 end
						   else
							 l)
						 q'
						 directions.(Random.int 16))
					  (if List.mem (i,j) t then
						begin
						  (* debug (Trace 10,Game) (sf "found (%d,%d) at distance %d" i j d); *)
						  ((i,j),d)::res
						end
					  else
						res)
					  r
				| [] -> loop (d + 1) q' res
			  in
			  loop2 [] res q
		in
		mark si sj;
		loop 0 [(si,sj)] []
  end

(* functions for displacement *)

let directions =
  [East,(0,1);
   West,(0,-1);
   South,(1,0);
   North,(-1,0)]

let square x = x *. x

let euclidian (i,j) (k,l) = sqrt (square (float (i - k)) +. square (float (j - l)))
let euclidian_squared (i,j) (k,l) = (square (float (i - k)) +. square (float (j - l)))
let euclidianf (i,j) (k,l) = sqrt (square (i -. k) +. square (j -. l))
let dirac x y = if x = y then 1.0 else 0.0
let manhattan i j i' j' = abs (i - i') + abs (j - j')

let client host port =
  debug (Info,Network) (sf "connecting to %S port %d" host port);
  let (ich,och) = open_connection (ADDR_INET((gethostbyname host).h_addr_list.(0), port)) in
  let xmit x =
	output_string och x;
	output_char och '\n';
	flush och
	(* debug (Trace 0,Network) (sf "transmitting %S" x) *)
  and recv () =
	let l = input_line ich in
	(* debug (Trace 0,Network) (sf "received %S" l); *)
	l
  in
  let xmit_command (bid,cmd) =
	xmit (sf "%d %s"
			bid
			(match cmd with
			  Move East -> "Move E"
			| Move West -> "Move W"
			| Move South -> "Move S"
			| Move North -> "Move N"
			| Pick pl -> "Pick "^(String.concat " " (List.map string_of_int pl))
			| Drop pl -> "Drop "^(String.concat " " (List.map string_of_int pl))))
  in
  (* Initialization - step 1 *)
  xmit "Player";
  (* Initialization - step 2 - read board *)
  (*  Transmitting the board - step 1 - board dimensions *)
  let (n',m') = split_at ' ' (recv ()) in
  let (n,m) = (int_of_string n',int_of_string m') in
  debug (Info,Game) (sf "board dimensions m = %d rows by n = %d columns" m n);
  (* brain-dead to coherent coordinates *)
  let i_of_y y = m - y
  and j_of_x x = x - 1
  and y_of_i i = m - i
  and x_of_j j = j + 1
  in
  (*  Transmitting the board - steps 2,... - board rows 1 to ..., from south to north, from west to east *)
  let a = Array.make_matrix m n Plain in
  let bases = ref IIM.empty in
  for i = 0 to m - 1 do
	let w = recv () in
	for j = 0 to n - 1 do
	  let (i',j') = (m - 1 - i,j) in
	  a.(m - 1 - i).(j) <-
		(match w.[j] with
		  '.' -> Plain
		| '~' -> Water
		| '#' -> Wall
		| '@' ->
			bases := IIM.add (i',j') { bi = i'; bj = j'; bpackages = Unknown } !bases;
			Base
		| _ ->
			debug (Critical,Game) (sf "invalid tile %C at 0-based position (%d,%d) in board ; assuming Plain" w.[j] i j);
			Plain)
	done
  done;
  let bases = !bases in
  debug (Info,Game) ("board read successfully");
  (* occupation update *)
  (* let occupation = Array.make_matrix m n 0.0 in
	 for i = 0 to m - 1 do
	 for j = 0 to n - 1 do
	 occupation.(i).(j) <-
	 match a.(i).(j) with
	 (Plain|Base) -> 0.0
	 | (Water|Wall) -> 1.0
	 done
	 done; *)
  (* compute dangerous points *)
  let eight_connex = Bitmap.make (m * n) false
  and near_water = Bitmap.make (m * n) false
  and pseudo_wall = Bitmap.make (m * n) false
  and pseudo_walls' = ref []
  in
  let get i j =
	if i >= 0 && i < m && j >= 0 && j < n then
	  a.(i).(j)
	else
	  Wall
  in
  let theoretically_walkable i j = match get i j with (Water|Wall) -> false | (Plain|Base) -> true in
  for i = 0 to m - 1 do
	for j = 0 to n - 1 do
	  Bitmap.set near_water (i * n + j)
		(get i (j - 1) = Water or
		 get i (j + 1) = Water or
		 get (i - 1) j = Water or
		 get (i + 1) j = Water);
	  Bitmap.set eight_connex (i * n + j)
		(theoretically_walkable (i - 1) (j - 1) &&
		 theoretically_walkable (i - 1) (j    ) &&
		 theoretically_walkable (i - 1) (j + 1) &&
		 theoretically_walkable (i    ) (j - 1) &&
		 theoretically_walkable (i    ) (j    ) &&
		 theoretically_walkable (i    ) (j + 1) &&
		 theoretically_walkable (i + 1) (j - 1) &&
		 theoretically_walkable (i + 1) (j    ) &&
		 theoretically_walkable (i + 1) (j + 1));
	  if
		theoretically_walkable i j &&
		((get (i + 1) j = Water &&
		  theoretically_walkable (i - 1) (j - 1) && 
		  theoretically_walkable (i - 1) (j) && 
		  theoretically_walkable (i - 1) (j + 1))
		   or
		 (get (i - 1) j = Water &&
		  theoretically_walkable (i + 1) (j - 1) && 
		  theoretically_walkable (i + 1) (j) && 
		  theoretically_walkable (i + 1) (j + 1))
		   or
		 (get i (j + 1) = Water &&
		  theoretically_walkable (i - 1) (j - 1) && 
		  theoretically_walkable i (j - 1) && 
		  theoretically_walkable (i + 1) (j - 1))
		   or
		 (get i (j - 1) = Water &&
		  theoretically_walkable (i - 1) (j + 1) && 
		  theoretically_walkable i (j + 1) && 
		  theoretically_walkable (i + 1) (j + 1)))
	  then
		begin
		  Bitmap.set pseudo_wall (i * n + j) true;
		  pseudo_walls' := (i,j)::!pseudo_walls'
		end
	done
  done;
  let pseudo_walls = !pseudo_walls' in
  (* display graphics *)
  let crt =
	if !Opt.crt then
	  Crt.init m n
	else
	  Crt.dummy_ops
  in
  if !Opt.crt then crt.Crt.redisplay a [];
  (* Initialization - step 3 - read player configuration *)
  let (id',w') = split_at ' ' (recv ()) in
  let (cc',fric') = split_at ' ' w' in
  let (id,cc,fric) = (int_of_string id', int_of_string cc', ref (int_of_string fric')) in
  debug (Info,Game) (sf "our id is %d (%S), our carrying capacity is %d (%S) and our amount of money is %d (%S)"
					   id id' cc cc' !fric fric');
  (* parse robots *)
  let b = Buffer.create 16 in
  let rec loop0 robots = function
	  [] -> robots (* no more robot positions *)
	| id'::x::x_pos'::y::y_pos'::r ->
		if id' = ""
			or String.length id' = 1
			or id'.[0] <> '#'
			or x <> "X"
			or y <> "Y"
		then
		  begin
			debug (Critical,Game) (sf "malformed robot info %S,%S,%S,%S,%S, stopping" id' x x_pos' y y_pos');
			robots
		  end
		else
		  let id = int_of_string (String.sub id' 1 (String.length id' - 1))
		  and x = int_of_string x_pos'
		  and y = int_of_string y_pos'
		  in
		  loop0 (IM.add id { id = id;
							 i = i_of_y y;
							 j = j_of_x x;
							 alive = true;
							 carrying = IS.empty;
							 history = Sliding.create !Opt.robot_history_length (-1,-1);
							 avoid = None
						   } robots) r
	| r ->
		debug (Critical,Game) (sf "robot positions -- number of elements not multiple of 5 -- rest is [%s], stopping"
								 (String.concat ";" (List.map (sf "%S") r)));
		robots
  in
  let robots = ref (loop0 IM.empty (split_multiple_at_char ' ' b (recv ()))) in
  let tracked_robots = ref IM.empty in (* robots being tracked... *)
  let packages = ref IM.empty in
  let score = ref 0 in
  let stick_u = ref 1.0
  and stick_v = ref 0.0
  and stick_counter = ref 0
  in
  let o = IM.fold (fun _ r count ->
	debug (Info,Game) (sf "robot number %d is at position (i = %d,j = %d)" r.id r.i r.j);
	count + 1) !robots 0
  in
  debug (Info,Game) (sf "total %d robot(s) including us" o);
  (* game play *)
  let us = IM.find id !robots in
  (* init breadth first search *)
  let bfs = Breadth_first.distances m n in
  let directions_scrambled =
	let a = Array.of_list directions in
	Array.init 16 (fun _ ->
	  scramble a;
	  Array.to_list a)
  in
  (* let palestine = ref IIS.empty in *)
  let repeat_predictor r =
	if Sliding.length r.history < 2 then
	  []
	else
	  let (i0,j0) = Sliding.get r.history 1 (* old *)
	  and (i1,j1) = Sliding.get r.history 0 (* new *)
	  in
	  let di = i1 - i0 and dj = j1 - j0 in
	  let rec loop i j k r =
		if k = !Opt.robot_history_length or not (theoretically_walkable i j) then
		  r
		else
		  loop (i + di) (j + dj) (k + 1) ((i,j)::r)
	  in
	  loop i1 j1 0 []
  in
  let predictors = [ repeat_predictor ] in
  let radius = !Opt.robot_radius in
  let robot_avoidance_turns = (m + n) / 16 in
  let angriness_level = ref 0 in
  let bid = ref 1 in
  let rec game_loop round =
	if !fric = 0 then
	  debug (Info,Game) ("no more money left, we are dead")
	else
	  (* get the list of packages available at current location *)
	  let rec loop1 pkgs = function
		  [] -> pkgs (* no more packages at this location *)
		| pid::px::py::pw::r ->
			(* no need to do unification because we are given everything we might need to know about this package *)
			loop1
			  (let pkg = { pid = pid;
						   ppickup = 0;
						   pi = us.i;
						   pj = us.j;
						   pdi = i_of_y py;
						   pdj = j_of_x px;
						   powner = None;
						   pw = pw }
			  in
			  packages := IM.add pid pkg !packages;
			  pkg::pkgs)
			  r
		| r ->
			debug (Critical,Game) (sf "packages -- number of elements not multiple of 4 -- rest is [%s], stopping"
									 (String.concat ";" (List.map (sf "%d") r)));
			pkgs
	  in
	  let pkgs_here = loop1 [] (List.map int_of_string (split_multiple_at_char ' ' b (recv ()))) in
	  (* debug (Trace 0,Game) (sf "money is %d units and we have %d packages at position (i = %d, j = %d)"
							  !fric
							  (List.length pkgs_here)
							  us.i us.j); *)
	  (* iterate over robots *)
	  let angriness = ref false in
	  let (robots_to_avoid',robots_alive') = IM.fold
		  (fun _ r (l,l') ->
			let d = manhattan us.i us.j r.i r.j in
			if r.id <> us.id && r.alive then
			  begin
				begin
				  if d <= 1 then
					begin
					(* contact with a robot *)
					if not !angriness then
					  incr angriness_level;
					angriness := true;
					bid := max 1 (min
									(1 + (Random.int !angriness_level))
									(!fric / 100));
					debug (Info,Strategy) (sf "fucking robot %d is blocking our way bidding %d" r.id !bid);
					if !angriness_level > !Opt.angriness_avoidance_threshold then
					  begin
						r.avoid <- Some round;
						debug (Info,Strategy) (sf "avoiding robot %d for %d turns..." r.id robot_avoidance_turns);
					  end;
				  end
				else
				  ();
				try
				  let t = IM.find r.id !tracked_robots in
				  if d > !Opt.robot_radius && round - t.last_track > !Opt.robot_tracking_persistence then
					begin
					  debug (Trace 7,Game) (sf "untracking robot %d" r.id);
					  tracked_robots := IM.remove r.id !tracked_robots
					end
				with
				  Not_found ->
					if d < !Opt.robot_radius then
					  begin
						debug (Trace 7,Game) (sf "robot %d entered tracking zone" r.id);
						tracked_robots := IM.add r.id
							{ robot = r;
							  predictors = List.map (fun f -> { score = 0.0; predictor = f }) predictors;
							  last_track = round } !tracked_robots;
					  end
					else
					  ();
				end;
				(match r.avoid with
				  None -> l
				| Some(x) ->
					if round - x > robot_avoidance_turns then
					  l
					else
					  (r.i,r.j)::l),(r.i,r.j)::l'
			  end
			else
			  l,l') !robots ([],[])
	  in
	  if !angriness then incr angriness_level else angriness_level := !angriness_level / 2;
	  (* let alpha = !Opt.occupation_update_rate
		 and occupation_threshold = !Opt.occupation_threshold
		 in *)
	  let walkable i j =
		match a.(i).(j) with
		  (Plain|Base) ->
			(Bitmap.get near_water (us.i * n + us.j) or not (Bitmap.get pseudo_wall (i * n + j))) &&
			not (List.mem (i,j) robots_to_avoid') &&
			not (List.mem (i,j) robots_alive' &&
				 manhattan us.i us.j i j < radius &&
				 Bitmap.get eight_connex (i * n + j))
		| (Water|Wall) -> false
	  in
	  let distances = bfs walkable
	  in
	  let go_towards i j =
		(* debug (Trace 0,Game) (sf "moving towards (%d,%d)" i j); *)
		if (us.i,us.j) = (i,j) then
		  begin
			(* debug (Warning,Game) (sf "we're already at (%d,%d) !" i j); *)
			(1,Drop [])
		  end
		else
		  let (_,cmd) =
			List.fold_left
			  (fun (best_dist,best_cmd) (dir,(di,dj)) ->
				let i',j' = us.i + di,us.j + dj in
				if 0 <= i' && i' < m && 0 <= j' && j' < n && walkable i' j' then
				  match distances (i',j') [(i,j)] with
					[] ->
					  (* point is not reachable *)
					  (* debug (Trace 10,Game) (sf "point is not reachable from (%d,%d)" i' j'); *)
					  (best_dist,best_cmd)
				  | (_,d)::_ -> if d < best_dist then (d,Move dir) else (best_dist,best_cmd)
				else
				  (best_dist,best_cmd))
			  (max_int,Drop [])
			  directions_scrambled.(Random.int (Array.length directions_scrambled))
		  in
		  (!bid,cmd)
	  in
	  (* convenient list of currently carried packages *)
	  let pkgs = IS.fold
		  (fun pid l ->
			try
			  (IM.find pid !packages)::l
			with
			  Not_found -> assert false (* l *))
		  us.carrying []
	  in
	  let service_carried_packages () =
		(* no packages are at their destination *)
		(* main heuristic here : cost / distance ratio *)
		(* compute, for each feasible direction, estimated gain *)
		(* select feasible direction maximizing gain *)
		incr stick_counter;
		let feasible_directions =
		  List.filter
			(fun (dir,(di,dj)) ->
			  let i',j' = us.i + di,us.j + dj in
			  0 <= i' && i' < m && 0 <= j' && j' < n && walkable i' j')
			directions
		in
		(* compute distances for packages and other robots *)
		(* debug (Trace 10,Strategy) (sf "stick %d (%f,%f)" !stick_counter !stick_u !stick_v); *)
		let costs =
		  List.map
			(fun (dir,(di,dj)) ->
			  let (i',j') = (us.i + di,us.j + dj) in
			  let pkgs_d = bfs walkable (i',j') (List.map (fun pkg -> (pkg.pdi,pkg.pdj)) pkgs) in
			  dir,
			  List.fold_left
				(fun score pkg ->
				  try
					let dist = float (List.assoc (pkg.pdi,pkg.pdj) pkgs_d) in (* distance of package *)
					if dist > 0.5 then
					  let desire = (float pkg.pw) /. (square dist) in
					  (* debug (Trace 10,Strategy) (sf "package %d destination (%d,%d) weight %d distance %f desire %f"
						 pkg.pid pkg.pdi pkg.pdj pkg.pw dist desire); *)
					  score +. desire
					else
					  max_float
				  with
					Not_found ->
					  (* debug (Trace 10,Game) (sf "package %d's destination (%d,%d) is unreachable" pkg.pid pkg.pdi pkg.pdj); *)
					  score)
				0.0
				pkgs)
			feasible_directions
		in
		(* dump costs *)
		let (_,cmd) = 
		  List.fold_left (fun (best_score,best_cmd) (dir,score) ->
			(* debug (Trace 0,Strategy)
			   (sf "direction %s score %f"
			   (match dir with East -> "E" | West -> "W" | South -> "S" | North -> "N")
			   score); *)
			if score > best_score then
			  (score,Move dir)
			else
			  (best_score,best_cmd)) (-.max_float,Drop []) costs
		in
		(1,cmd)
	  in
	  (* if this is a base, update the known number of packages at it... *)
	  begin
		try
		  let b = IIM.find (us.i,us.j) bases in
		  b.bpackages <- Known (List.length pkgs_here)
		with
		  Not_found -> ()
	  end;
	  let bid,cmd =
		if pkgs_here = [] then
		  begin
			(* no packages, is this a base ? *)
			(* are we carrying any packages ? *)
			
			if IS.is_empty us.carrying then
			  (* we don't own any package, go to a base... *)
			  (let b' =
				match
				  IIM.fold (fun _ b l ->
					match b.bpackages with
					  Known 0 -> l (* we know there's nothing at that base... *)
					| (Unknown|Known _) -> (b.bi,b.bj)::l) bases []
				with
				  [] -> IIM.fold (fun _ b l -> (b.bi,b.bj)::l) bases []
				| b' -> b'
			  in
			  match b' with
				[] ->
				  (* no bases !!!! randomly walk... *)
				  (* debug (Trace 7,Game) ("no bases !"); *)
				  go_towards (Random.int m) (Random.int n)
			  | _ ->
				  (* debug (Trace 7,Game) (sf "at least some bases : [%s]" (String.concat ";" (List.map (fun (i,j) -> sf "%d,%d" i j) b'))); *)
				  let d = distances (us.i,us.j) b' in
				  (* find base with maximum package / distance ratio *)
				  let ((i,j),d) =
					List.fold_left
					  (fun ((ci,cj),cd) ((i,j),d) ->
						(* debug (Trace 7,Game) (sf "base at %d %d with distance %d" i j d); *)
						if d < cd or d = cd && i < ci or d = cd && i = ci && j < cj then
						  ((i,j),d)
						else
						  ((ci,cj),cd)) ((-1,-1),max_int) d
				  in
				  if i = -1 then
					begin
					  (* debug (Trace 7,Game) (sf "no reachable bases !?"); *)
					  go_towards (Random.int m) (Random.int n)
					end
				  else
					go_towards i j)
			else
			  (* we are carrying some packages, check if any of them is at it's destination *)
			  let to_drop =
				List.filter (fun pkg ->
				  if (pkg.pdi,pkg.pdj) = (us.i,us.j) then
					begin
					  score := !score + pkg.pw;
					  debug (Info,Game) (sf "package %d arrived at destination ; new score = %d" pkg.pid !score);
					  if !Opt.crt then
						  ignore (crt.Crt.create pkg.pdi pkg.pdj Crt.Delivery 2.0);
					  true
					end
				  else
					false) pkgs
			  in
			  match to_drop with
				[] ->
				  (* debug (Trace 0,Game) ("no packages arrived at destination, servicing carried ones"); *)
				  service_carried_packages ()
			  | _ ->
				  (* dropping -- reset self escape history *)
				  Sliding.clear us.history;
				  (* reset stick *)
				  stick_counter := 0;
				  let theta = Random.float two_pi in
				  stick_u := cos theta;
				  stick_v := sin theta;
				  (1,Drop (List.map (fun pkg -> pkg.pid) to_drop))
		  end
		else
		  (* we do have some packages here *)
		  (* compute spare capacity *)
		  let carried = IS.fold (fun pid w -> w + try (IM.find pid !packages).pw with Not_found -> assert false) us.carrying 0 in
		  let spare = cc - carried
		  in
		  (* debug (Trace 0,Game) (sf "some packages here ; carrying %d ; spare %d" carried spare); *)
		  let ph = Array.of_list pkgs_here in
		  let rec loop
			  i
			  best_pkgs
			  best_weight
			  =
			if i = !Opt.max_package_tries then
			  Array.to_list (Array.map (fun p -> p.pid) best_pkgs)
			else
			  (* randomized greedy algorithm *)
			  begin
				scramble ph;
				let rec randomized total_weight j =
				  if j = Array.length ph or total_weight + ph.(j).pw > spare then
					(j,total_weight)
				  else
					randomized (total_weight + ph.(j).pw) (j + 1)
				in
				let (j,total_weight) = randomized 0 0 in
				if total_weight > best_weight then
				  loop (i + 1) (Array.sub ph 0 j) total_weight
				else
				  loop (i + 1) best_pkgs best_weight
			  end
		  in
		  match loop 0 [||] (-1) with
			[] ->
			  (* no packages to pickup *)
			  (* service carried packages instead *)
			  service_carried_packages ()
		  | pick_me ->
			  Sliding.clear us.history;
			  (1,Pick pick_me)
	  in
	  xmit_command (bid,cmd);
	  fric := !fric - (abs bid);
	  (* what really happened ? *)
	  (* #5 #4 E *)
	  let rec loop2 wrh = function
		  [] -> wrh
		| rid'::r ->
			if rid' = "" or rid'.[0] <> '#' then
			  begin
				debug (Critical,Game) (sf "malformed robot id %S, stopping" rid');
				wrh
			  end
			else
			  let rid = int_of_string (String.sub rid' 1 (String.length rid' - 1)) in
			  loop3 wrh rid [] r
	  and loop3 wrh rid things = function
		  [] -> (rid,List.rev things)::wrh
		| "E"::r -> loop3 wrh rid (Moved East::things) r
		| "W"::r -> loop3 wrh rid (Moved West::things) r
		| "N"::r -> loop3 wrh rid (Moved North::things) r
		| "S"::r -> loop3 wrh rid (Moved South::things) r
		| "P"::pid'::r -> loop3 wrh rid ((Picked(int_of_string pid'))::things) r
		| "D"::pid'::r -> loop3 wrh rid ((Dropped(int_of_string pid'))::things) r
		| "X"::x'::"Y"::y'::r ->
			loop3 wrh rid ((Appeared(i_of_y (int_of_string y'),
									 j_of_x (int_of_string x')))::things) r
		| (cmd'::r) as r' ->
			if cmd' <> ""  && cmd'.[0] = '#' then
			  loop2 ((rid,List.rev things)::wrh) r'
			else
			  begin
				debug (Critical,Game) (sf "unknown command %S, stopping" cmd');
				wrh
			  end
	  in
	  let wrh = loop2 [] (split_multiple_at_char ' ' b (recv ())) in

	  (* mark all robots as dead *)
	  IM.iter (fun _ r -> r.alive <- false) !robots;

	  (* process what really happened *)
	  List.iter (function
		  (rid,things) ->
			if not (IM.mem rid !robots) then
			  begin
				List.iter (function
				  | Appeared(i,j) ->
					  debug (Info,Game) (sf "we have a new robot with id %d at (%d,%d)" rid i j);
					  robots := IM.add rid { id = rid;
											 i = i;
											 j = j;
											 alive = true;
											 carrying = IS.empty;
											 history = Sliding.create !Opt.robot_history_length (-1,-1);
											 avoid = None
										   } !robots
				  | _ -> ()) things
			  end;
			try
			  let r = IM.find rid !robots in
			  r.alive <- true;
			  (* if not r.alive then
				 debug (Info,Game) (sf "robot %d is a zombie !" rid); *)
			  let update i j =
				if i < 0 or i >= m or j < 0 or j > n or a.(i).(j) = Wall then
				  debug (Info,Game) (sf "robot %d attempted to move to wall or beyond map, ignoring" rid)
				else
				  begin
					r.i <- i;
					r.j <- j;
					if r.alive && a.(i).(j) = Water then
					  begin
						debug (Info,Game) (sf "robot %d drowned" rid);
						r.alive <- false
					  end
				  end
			  in
			  List.iter (function
				  Moved East -> update r.i (r.j + 1)
				| Moved West -> update r.i (r.j - 1)
				| Moved North -> update (r.i - 1) r.j (* BRAIN-DAMAGED COORDINATES *)
				| Moved South -> update (r.i + 1) r.j (* BRAIN-DAMAGED COORDINATES *)
				| Appeared(i,j) -> update i j
				| Dropped pid -> r.carrying <- IS.remove pid r.carrying
				| Picked pid ->
					if rid = us.id then
					  begin
						try
						  let pkg = IM.find pid !packages in
						  pkg.ppickup <- round
						with
						  Not_found -> ();
					  end;
					r.carrying <- IS.add pid r.carrying) things
			with
			  Not_found ->
				debug (Critical,Game) (sf "unknown robot %d, skipping" rid)) wrh;
	  IM.iter (fun _ r ->
		(* if r.id <> us.id then
		   begin
		   let x = occupation.(r.i).(r.j) in
		   let y = alpha *. x +. (1.0 -. alpha) in
		   occupation.(r.i).(r.j) <- y;
		   begin
		   match x > occupation_threshold,y > occupation_threshold with
		   (false,false|true,true) -> ()
		   | false,true -> palestine := IIS.add (r.i,r.j) !palestine;
		   | true,false -> palestine := IIS.remove (r.i,r.j) !palestine;
		   end;
		   end; *)
		Sliding.push r.history (r.i,r.j)) !robots;

	  (* display *)
	  if !Opt.crt then
		begin
		  crt.Crt.redisplay a
			((List.map (fun (i,j) -> (i,j,Crt.Pseudo_wall)) pseudo_walls)@
			 (List.map (fun pkg -> (pkg.pi,pkg.pj,Crt.Package pkg.pid)) pkgs)@
			 (List.map (fun pkg -> (pkg.pdi,pkg.pdj,Crt.Target pkg.pid)) pkgs)@
			 (IM.fold (fun _ r list -> Sliding.fold (fun list (i,j) -> (i,j,Crt.Trail)::list) list r.history) !robots [])@
			 (* (IIS.fold (fun (i,j) list ->
				occupation.(i).(j) <- alpha *. occupation.(i).(j);
				(i,j,Crt.Occupied)::list) !palestine [])@ *)
			 (IM.fold
				(fun _ t l -> 
				  List.fold_left (fun l p ->
					let rec loop i r =
					  if i = radius then
						r
					  else
						loop1 (i + 1) (- radius) r
					and loop1 i j r =
					  if j = radius then
						loop (i + 1) r
					  else
						if List.mem (i,j) (p.predictor t.robot) then
						  loop1 i (j + 1) ((i,j,Crt.Prediction)::r)
						else
						  loop1 i (j + 1) r
					in
					loop 0 l) ((t.robot.i,t.robot.j,Crt.Track)::l) t.predictors)
				!tracked_robots
				(IM.fold
				   (fun _ r l ->
					 if r.id <> us.id then
					   (r.i,r.j,if r.alive then Crt.Robot r.id else Crt.Dead_robot r.id)::l
					 else
					   l) !robots 
				   [(us.i,us.j,Crt.Myself)])));
		end;
	  (* play *)
	  if us.alive then
		game_loop (round + 1)
	  else
		debug (Info,Game) (sf "we just died");
  in
  game_loop 0

let _ =
  debug_enabled := []; (* Only[Info;Critical;Warning],All]; *)
  Random.self_init ();
  try
	let host = ref None
	and port = ref None
	in
	Arg.parse
	  Opt.specs
	  (fun x ->
		match !host with
		  None -> host := Some x
		| Some _ -> 
			match !port with
			  None -> port := Some (int_of_string x)
			|Some _ -> debug (Warning,System) (sf "extra argument %S ignored" x))
	  (sf "usage: %s hostname port" Sys.argv.(1));
	match (!host,!port) with
	  Some host, Some port -> client host port
	| _,_ -> failwith "host or port is missing"
  with
	Unix_error(err,u,v) ->
	  debug (Critical,Game) (sf "unix error %s with arguments %S and %S" (error_message err) u v)
  | x ->
	  debug (Critical,Game) (sf "uncaught exception %s" (Printexc.to_string x))


